////////////////////////////////////////////////////////////////////////////////////////
// File: scripts/main.js
// Usage: main scripts functionality
////////////////////////////////////////////////////////////////////////////////////////

// main imports
import { system, world } from "@minecraft/server";

// features
import { injectHandler } from "./injects/injects.js";
import * as explo from "./custom_explosion.js";
import * as stun from "./stun.js";
import { specialAttackHandler } from "./item_components.js";
import { goatAbilityHandler } from "./weapons/usable.js";
import { tornadoMovement } from "./weapons/melee.js";
import { randomTeleport, wardenSonicBlaster1, wardenSonicBlaster2, wardenSonicBlaster3 } from "./weapons/ranged.js";
import { getEntitiesFromRadius } from "./weapons/util.js";
import * as onHit from "./weapons/on_hit_entity.js";

////////////////////////////////////////////////////////////////////////////////////////

// run every tick
system.runInterval(() => {

  // Update player list
  const players = world.getAllPlayers();

  // execute as all players
  for (const player of players) {
    stun.playerStunHandler(player); // handle stunned players
    goatAbilityHandler(player) // handle goating players
  }
});

////////////////////////////////////////////////////////////////////////////////////////

// script events
const eventHandlers = {
  "sbs_mw:inject": (event) => injectHandler(event),
  "sbs_mw:create_explosion": (event) => explo.createCustomExplosion(event),
  "sbs_mw:explosion_knockback": (event) => explo.explosionKnockback(event),
  "sbs_mw:stun": (event) => stun.stunEventHandler(event),
  "sbs_mw:tornado_move": (event) => tornadoMovement(event),
  "sbs_mw:random_tp": (event) => randomTeleport(event),
};
system.afterEvents.scriptEventReceive.subscribe((event) => {
  for (const [prefix, handler] of Object.entries(eventHandlers)) {
    if (event.id.startsWith(prefix)) {
      handler(event); // Call the handler if the prefix matches
      break;
    }
  }
});

////////////////////////////////////////////////////////////////////////////////////////

// Define a lookup object for projectiles and their corresponding hit handler functions
const projectileHandlers = {
  // MELEE PROJECTILES
  "sbs_mw:cave_spider_sword_projectile": onHit.caveSpiderSword,
  "sbs_mw:drowned_trident_projectile": onHit.drownedTrident,
  "sbs_mw:vindicator_axe_projectile": onHit.vindicatorAxe,

  // RANGED PROJECTILES
  "sbs_mw:blaze_flamethrower_projectile": onHit.blazeFlamethrower,
  "sbs_mw:bogged_bow_projectile": onHit.boggedBow,
  "sbs_mw:creeper_cannon_projectile": onHit.creeperCannon,
  "sbs_mw:enderman_phaser_projectile": onHit.endermanPhaser,
  "sbs_mw:ghast_blaster_projectile": onHit.ghastBlaster,
  "sbs_mw:llama_spitter_projectile": onHit.llamaSpitter,
  "sbs_mw:magma_flodder_projectile": onHit.magmaFlooder,
  "sbs_mw:pillager_crossbow_projectile": onHit.pillagerCrossbow,
  "sbs_mw:skeleton_bow_projectile": onHit.skeletonBow,
  "sbs_mw:spider_net_projectile": onHit.spiderNet,
  "sbs_mw:stray_bow_projectile": onHit.strayBow,
  "sbs_mw:witch_wand_projectile": onHit.witchWand,
  "sbs_mw:warden_sonic_blaster_projectile1": onHit.wardenSonicBlaster,
  "sbs_mw:warden_sonic_blaster_projectile2": onHit.wardenSonicBlaster,
  "sbs_mw:warden_sonic_blaster_projectile3": onHit.wardenSonicBlaster,

  // Throwable PROJECTILES
  "sbs_mw:chicken_bomb_projectile": onHit.chickenBomb,
  "sbs_mw:pig_mudder_projectile": onHit.pigMudder,
  "sbs_mw:slime_jar_projectile": onHit.slimeJar
};

// Projectile hit event handler
world.afterEvents.projectileHitEntity.subscribe((event) => {
  const handler = projectileHandlers[event.projectile.typeId];
  if (handler) {
    handler(event);
  }
});

////////////////////////////////////////////////////////////////////////////////////////

world.afterEvents.projectileHitBlock.subscribe((event) => {
  switch (event.projectile.typeId) {
    case "sbs_mw:warden_sonic_blaster_projectile1":
    case "sbs_mw:warden_sonic_blaster_projectile2":
    case "sbs_mw:warden_sonic_blaster_projectile3":
      const hitBlock = event.getBlockHit()?.block;
      const strengthMapping = {
        "sbs_mw:warden_sonic_blaster_projectile1": 2,
        "sbs_mw:warden_sonic_blaster_projectile2": 4,
        "sbs_mw:warden_sonic_blaster_projectile3": 6,
      };
      const verticalMultiplierMapping = {
        "sbs_mw:warden_sonic_blaster_projectile1": 1,
        "sbs_mw:warden_sonic_blaster_projectile2": 2,
        "sbs_mw:warden_sonic_blaster_projectile3": 3,
      };

      const strength = strengthMapping[event.projectile.typeId];
      const verticalMultiplier = verticalMultiplierMapping[event.projectile.typeId];

      for (const entity of getEntitiesFromRadius(hitBlock, 8)) {
        try {
          const knockbackStrength = event.hitVector.x * event.hitVector.z * 5.45 + strength;
          const verticalStrength = event.hitVector.y * 1.15 + verticalMultiplier;
          entity.applyKnockback(event.hitVector.x, event.hitVector.z, knockbackStrength, verticalStrength);
        } catch (err) {
          console.log(`KNOCKBACK ERROR :: ${err}`);
        }
      }
      break;
    default:
      break;
  }
});

////////////////////////////////////////////////////////////////////////////////////////

// custom item components
world.beforeEvents.worldInitialize.subscribe(({ itemComponentRegistry }) => {
  itemComponentRegistry.registerCustomComponent("sbs_mw:special_attack_handler", specialAttackHandler)
});

////////////////////////////////////////////////////////////////////////////////////////

world.afterEvents.itemReleaseUse.subscribe((event) => {
  const { itemStack, source, useDuration } = event;
  // warden sonic blaster
  const useTick = 9999 * 20 - useDuration
  if (source.typeId === "minecraft:player" && itemStack.typeId === "sbs_mw:warden_sonic_blaster") {
    const durationToBlaster = [
      { min: 0, max: 20, blaster: wardenSonicBlaster1 },
      { min: 20, max: 40, blaster: wardenSonicBlaster2 },
      { min: 40, max: 60, blaster: wardenSonicBlaster3 },
    ];
    for (const { min, max, blaster } of durationToBlaster) {
      if (useTick > min && useTick <= max) {
        blaster(event);
        break;
      }
    }
    if (useTick > 60) { // Default to wardenSonicBlaster3 if useTick > 60
      wardenSonicBlaster3(event);
    }
  }
});

////////////////////////////////////////////////////////////////////////////////////////
